"""Test Target resolution."""
from __future__ import absolute_import, unicode_literals

from travis_log_fetch.config import _get_travispy

from travis_log_fetch.get import get_user_repos

import pytest


class TestUser(object):

    def test_travispy(self):
        _travis = _get_travispy()

        repos = get_user_repos(_travis, 'travispy')

        slugs = [repo.slug for repo in repos]
        assert 'travispy/on_pypy' in slugs



"""Test loading historical builds and jobs."""
from __future__ import absolute_import, unicode_literals

from travis_log_fetch.config import _get_travispy

from travis_log_fetch._target import Target
from travis_log_fetch.get import (
    get_travis_repo,
    get_historical_builds,
    get_historical_build,
    get_historical_job,
)

import pytest


class TestHistorical(object):

    def test_latest(self):
        _travis = _get_travispy()
        repo = get_travis_repo(_travis, 'travispy/on_pypy')

        builds = get_historical_builds(_travis, repo)
        build = next(builds)
        assert build.repository_id == 2598880
        assert build.id == repo.last_build_id

    def test_after(self):
        _travis = _get_travispy()
        repo = get_travis_repo(_travis, 'travispy/on_pypy')

        builds = get_historical_builds(_travis, repo,
                                       _after=3, _load_jobs=False)
        build = next(builds)
        assert build.repository_id == 2598880
        assert build.number == '2'

        build = next(builds)
        assert build.repository_id == 2598880
        assert build.number == '1'

    def test_all_small(self):
        _travis = _get_travispy()
        repo = get_travis_repo(_travis, 'travispy/on_pypy')

        builds = get_historical_builds(_travis, repo)

        ids = []
        for build in builds:
            assert build.repository_id == 2598880
            ids.append(build.id)

        assert ids == [53686685, 37521698, 28881355]

    def test_multiple_batches_menegazzo(self):
        """Test using a repository that has greater than 2*25 builds."""
        # Ideally each has one or two jobs, so that doesnt slow down the test,
        # and the logs are small in case the log is fetched with the job.
        _travis = _get_travispy()
        repo = get_travis_repo(_travis, 'menegazzo/travispy')

        builds = get_historical_builds(_travis, repo, _load_jobs=False)

        ids = []
        prev_number = None
        for build in builds:
            assert build.repository_id == 2419489

            if int(build.number) in [80]:
                # There are two '80'
                # See https://github.com/travis-ci/travis-ci/issues/2582
                print('duplicate build number {0}: {1}'.format(
                    build.number, build.id))

                assert build.id in [45019395, 45019396]
                if build.id == 45019395:
                    assert prev_number == int(build.number)
                else:
                    assert prev_number == int(build.number) + 1
            elif prev_number:
                # All other build numbers decrease rather orderly
                assert prev_number == int(build.number) + 1

            prev_number = int(build.number)

            if ids:
                assert build.id < ids[-1]

            ids.append(build.id)
            if len(ids) > 100:
                break

        assert len(ids) == len(set(ids))

    def test_multiple_batches_bootstrap(self):
        """Test using a repository that has lots of builds, esp. PRs."""
        _travis = _get_travispy()
        repo = get_travis_repo(_travis, 'twbs/bootstrap')

        builds = get_historical_builds(_travis, repo,
                                       _after=12071,
                                       _load_jobs=False)

        ids = []
        prev_number = None
        for build in builds:
            assert build.repository_id == 12962

            if int(build.number) in [12069, 12062, 12061, 12054, 12049,
                                     12048, 12041, 12038, 12037, 12033]:
                # Many duplicates
                # See https://github.com/travis-ci/travis-ci/issues/2582
                print('duplicate build number {0}: {1}'.format(
                    build.number, build.id))

                if build.id in [53437234, 53350534, 53350026,
                                53263731, 53263730,  # two extra 12054
                                53180440, 53179846, 53062896, 53019568,
                                53004896, 52960766]:
                    assert prev_number == int(build.number)
                else:
                    assert prev_number == int(build.number) + 1

            elif prev_number:
                # All other build numbers decrease rather orderly
                assert prev_number == int(build.number) + 1

            prev_number = int(build.number)

            if ids:
                assert build.id < ids[-1]

            ids.append(build.id)
            # There are many more duplicates, so we stop here.
            if int(build.number) == 12033:
                break

        assert len(ids) == len(set(ids))

    def test_logical_single_job_build(self):
        target = Target.from_extended_slug('travispy/on_pypy#1.1')

        _travis = _get_travispy()
        job = get_historical_job(_travis, target)

        assert job.repository_id == 2598880
        assert job.number == '1.1'
        assert job.id == 28881356

    def test_logical_multiple_job_build(self):
        target = Target.from_extended_slug('menegazzo/travispy#101.3')

        _travis = _get_travispy()
        job = get_historical_job(_travis, target)

        assert job.repository_id == 2419489
        assert job.number == '101.3'
        assert job.id == 82131391

    def test_logical_duplicate_build(self):
        target = Target.from_extended_slug('menegazzo/travispy#80.3')

        _travis = _get_travispy()
        pytest.raises(AssertionError, get_historical_build, _travis, target)



# -*- coding: utf-8 -*-
"""Configuration module."""
from __future__ import absolute_import
from __future__ import unicode_literals

import sys

from logging import getLogger
from os.path import expanduser
from textwrap import dedent

import configargparse

import github3

import travispy

_options = None
_travispy = None
_github = None

__logs__ = getLogger(__package__)


def get_parser():
    """Get config parser."""
    parser = configargparse.ArgParser(
        prog='travis_log_fetch',
        usage='%(prog)s [options] [target [target2 ...]]',
        description=dedent(
            """
            travis_log_fetch fetches many log files.
            """.rstrip()),
        default_config_files=[
            '~/.travisrc', '.travisrc'])

    parser.add('-c', '--config', is_config_file=True,
               help='config file path')
    parser.add('-d', '--dir', help='log storage directory',
               default='~/.travis')
    parser.add('--access-token', required=False, help='Github access token',
               env_var='GITHUB_ACCESS_TOKEN')
    parser.add('-v', '--verbose', help='verbose', action='store_true')
    parser.add('-r', '--refresh', help='refresh', action='store_true')
    parser.add('--forks', help='fetch forks', action='store_true')
    parser.add('-f', '--force', help='force', action='store_true')
    parser.add('--format', help='log filename format',
               default='{job.repository.slug}/{job.number}-{job.state}.txt')
    parser.add('-a', '--all', help='all', action='store_true')
    parser.add('-o', '--old', help='old', action='store_true')
    parser.add('-s', '--self', help='fetch own repos', action='store_true')
    parser.add('-w', '--wait', help='wait for jobs to complete',
               action='store_true')
    parser.add('--sleep', help='time to wait for jobs to complete',
               type=int, default=30)
    parser.add('--count', help='number of old logs to fetch',
               type=int, default=10)
    parser.add('targets', nargs='*', help='targets')

    return parser


def get_options():
    """Get options."""
    global _options

    if not _options:
        parser = get_parser()
        args = None
        if sys.argv[0].endswith('/py.test'):
            args = []
        try:
            _options = parser.parse_args(args=args)
            __logs__.debug('config options: {0}'.format(_options))
        except SystemExit:
            __logs__.warning('failed to load options')
            return

        _options.dir = expanduser(_options.dir)

    return _options


def _get_travispy():
    """Get travis-ci handle."""
    global _travispy

    if not _travispy:
        options = get_options()
        if options and options.access_token:
            _travispy = travispy.TravisPy.github_auth(options.access_token)
            __logs__.debug('logged into travis-ci')
        else:
            _travispy = travispy.TravisPy()
            __logs__.debug('anon travis-ci activated')

    return _travispy


def _get_github():
    """Get github handle."""
    global _github

    if not _github:
        options = get_options()
        if options and options.access_token:
            _github = github3.login(token=options.access_token)
            __logs__.debug('logged into github')

        if not _github:
            _github = github3.GitHub()
            __logs__.debug('anon github activated')

    return _github



# -*- coding: utf-8 -*-
"""Object that identifies a target resource."""
from __future__ import absolute_import
from __future__ import unicode_literals

from travispy import Build, Job, Repo


class Target(object):
    """Identifier of a target resource."""

    def __init__(self, user=None, project=None, build_id=None, job_id=None,
                 build_number=None, job_number=None, slug=None):
        """
        Constructor.

        build_id and job_id are the real identifier of those objects.
        build_number and job_number are the relative logical numbering
        of each object within their container.
        e.g. within a repo, a logical number '10.2' is the second job
        of the tenth build.
        """
        self.user = user
        self.project = project
        if slug:
            self.slug = slug

        types = [
            type(obj) for obj in [build_id, job_id, build_number, job_number]]

        assert set(types) <= set([type(None), int]), \
            'found types {0}'.format(types)

        self.build_id = build_id
        self.job_id = job_id
        self.build_number = build_number
        self.job_number = job_number
        self.state = None

    @property
    def slug(self):
        """Get repository slug."""
        assert self.user and self.project
        return self.user + '/' + self.project

    @slug.setter
    def slug(self, value):
        """Set repository slug."""
        parts = value.split('/', 1)
        assert len(parts) == 2
        assert '/' not in parts[1]
        self.user = parts[0]
        self.project = parts[1]

    @property
    def number(self):
        """Get logical number comprised of build and job number."""
        assert self.build_number
        assert self.job_number
        return '{0}.{1}'.format(self.build_number, self.job_number)

    @number.setter
    def number(self, value):
        """Set logical number."""
        assert '/' not in value
        parts = value.split('.', 1)
        assert len(parts) == 2
        assert '.' not in parts[1]
        self.build_number = int(parts[0])
        self.job_number = int(parts[1])

    @property
    def extended_slug(self):
        """Create an extended slug such as <slug>/10.2."""
        if self.build_number and self.job_number:
            return '{0}/{1}'.format(self.slug, self.number)
        elif self.build_number:
            return '{0}/{1}'.format(self.slug, self.build_number)
        elif self.job_id:
            return '{0}:{1}'.format(self.slug, self.job_id)
        elif self.build_id:
            return '{0}@{1}'.format(self.slug, self.build_id)
        else:
            return self.slug

    def __str__(self):
        """String representation using extended slug."""
        if not self.user or not self.project:
            return '<invalid>'
        return self.extended_slug

    def __repr__(self):
        """Internal representation using extended slug."""
        if not self.user or not self.project:
            return '<invalid>'
        return '<{0}: {1}>'.format(self.__class__.__name__, self.extended_slug)

    def __eq__(self, other):
        """Compare to other for equivalence."""
        return other == self.extended_slug

    @classmethod
    def from_simple_slug(cls, slug):
        """Return an Target from a slug containing user and project."""
        parts = slug.rsplit('/', 1)
        assert len(parts) == 2
        assert '/' not in parts[1]
        obj = cls(*parts)
        return obj

    @classmethod
    def from_extended_slug(cls, slug):
        """Return an Target from an build or job slug."""
        parts = slug.split('/')

        assert len(parts) in [2, 3]

        user = parts[0]

        logical_id = None
        build_id = job_id = None

        if len(parts) == 3:
            project, logical_id = parts[1:]
        else:
            if '#' in parts[1]:
                project, logical_id = parts[1].split('#', 1)
            elif '@' in parts[1]:
                project, build_id = parts[1].split('@', 1)
                build_id = int(build_id)
            elif ':' in parts[1]:
                project, job_id = parts[1].split(':', 1)
                job_id = int(job_id)
            else:
                project = parts[1]

        if logical_id:
            build_number, sep, job_number = logical_id.partition('.')
            assert(build_number)
            build_number = int(build_number)
            if job_number:
                job_number = int(job_number)
            else:
                job_number = None

            # print('{0} . {1}'.format(type(build_number), type(job_number)))

            return cls(user, project,
                       build_number=build_number, job_number=job_number)

        return cls(user, project, build_id, job_id)

    @classmethod
    def from_url(cls, url):
        """Create an Target from a travis url."""
        _, _, path = url.partition('//travis-ci.org/')
        assert path
        path, _, line_no = path.partition('#')

        parts = path.split('/')

        if len(parts) == 2:
            return cls(parts[0], parts[1])

        if parts[2] == 'jobs':
            job_id = int(parts[3])
            return cls(parts[0], parts[1], None, job_id)
        elif parts[2] == 'builds':
            build_id = int(parts[3])
            return cls(parts[0], parts[1], build_id, None)
        else:
            raise AssertionError('unknown url {0}'.format(url))

    @classmethod
    def _from_travispy_obj(cls, obj):
        if isinstance(obj, Repo):
            return cls.from_simple_slug(obj.slug)
        elif isinstance(obj, Build):
            return cls(slug=obj.repository.slug, build_id=obj.id,
                       build_number=int(obj.number))
        elif isinstance(obj, Job):
            _obj = cls(job_id=obj.id, slug=obj.slug)
            _obj.number = obj.number
            return _obj
        else:
            raise AssertionError('unknown object {0!r}'.format(obj))



# -*- coding: utf-8 -*-
"""Fetch travis logs."""
from __future__ import absolute_import
from __future__ import unicode_literals

from travis_log_fetch.cmdline import main

__version__ = '0.1.0'

main



"""Test Target class."""
from __future__ import absolute_import, unicode_literals

from travis_log_fetch._target import Target

import pytest


class Test(object):

    def test_empty(self):
        target = Target()
        assert target.user is None
        assert target.project is None
        assert target.build_id is None
        assert target.job_id is None
        assert target.build_number is None
        assert target.job_number is None

        #pytest.raises(AssertionError, getattr, target, 'slug')
        #pytest.raises(AssertionError, getattr, target, 'extended_slug')

    def test_invalid_basic_slug(self):
        pytest.raises(AssertionError, Target.from_simple_slug, 'foobar')

    def test_basic_slug(self):
        target = Target.from_simple_slug('foo/bar')
        assert target.user == 'foo'
        assert target.project == 'bar'
        assert target.slug == 'foo/bar'
        assert target.build_id is None
        assert target.job_id is None
        assert target.build_number is None
        assert target.job_number is None


class TestExtended(object):

    def test_slash(self):
        target = Target.from_extended_slug('foo/bar/10.1')
        assert target.user == 'foo'
        assert target.project == 'bar'
        assert target.slug == 'foo/bar'
        assert target.build_id is None
        assert target.job_id is None
        assert target.build_number == 10
        assert target.job_number == 1

    def test_hash(self):
        target = Target.from_extended_slug('foo/bar#10.1')
        assert target.user == 'foo'
        assert target.project == 'bar'
        assert target.slug == 'foo/bar'
        assert target.build_id is None
        assert target.job_id is None
        assert target.build_number == 10
        assert target.job_number == 1

    def test_at(self):
        target = Target.from_extended_slug('foo/bar@9999')
        assert target.user == 'foo'
        assert target.project == 'bar'
        assert target.slug == 'foo/bar'
        assert target.build_id == 9999
        assert target.job_id is None
        assert target.build_number is None
        assert target.job_number is None

    def test_colon(self):
        target = Target.from_extended_slug('foo/bar:10000')
        assert target.user == 'foo'
        assert target.project == 'bar'
        assert target.slug == 'foo/bar'
        assert target.build_id is None
        assert target.job_id == 10000
        assert target.build_number is None
        assert target.job_number is None

class TestUrl(object):

    def test_https_project_url(self):
        target = Target.from_url('https://travis-ci.org/foo/bar')
        assert target.user == 'foo'
        assert target.project == 'bar'
        assert target.slug == 'foo/bar'
        assert target.build_id is None
        assert target.job_id is None
        assert target.build_number is None
        assert target.job_number is None

    def test_http_project_url(self):
        target = Target.from_url('http://travis-ci.org/foo/bar')
        assert target.user == 'foo'
        assert target.project == 'bar'
        assert target.slug == 'foo/bar'
        assert target.build_id is None
        assert target.job_id is None
        assert target.build_number is None
        assert target.job_number is None

    def test_no_protocol_project_url(self):
        target = Target.from_url('//travis-ci.org/foo/bar')
        assert target.user == 'foo'
        assert target.project == 'bar'
        assert target.slug == 'foo/bar'
        assert target.build_id is None
        assert target.job_id is None
        assert target.build_number is None
        assert target.job_number is None

    def test_build_url(self):
        target = Target.from_url('//travis-ci.org/foo/bar/builds/9999')
        assert target.user == 'foo'
        assert target.project == 'bar'
        assert target.slug == 'foo/bar'
        assert target.build_id == 9999
        assert target.job_id is None
        assert target.build_number is None
        assert target.job_number is None

    def test_job_url(self):
        target = Target.from_url('//travis-ci.org/foo/bar/jobs/10000')
        assert target.user == 'foo'
        assert target.project == 'bar'
        assert target.slug == 'foo/bar'
        assert target.build_id is None
        assert target.job_id == 10000
        assert target.build_number is None
        assert target.job_number is None



"""Test log store."""
from __future__ import absolute_import, unicode_literals

from travis_log_fetch._store import (
    get_repo_stored_builds,
    get_stored_repo_slugs,
    get_stored_targets,
)

import pytest


class TestDefaultLayout(object):

    base_dir = './tests/default_layout'
    layout = '{job.repository.slug}/{job.number}-{job.state}.txt'

    def test_slugs(self):
        slugs = get_stored_repo_slugs(self.base_dir, self.layout)

        assert 'wikimedia/pywikibot-core' in slugs
        assert 'foo' not in slugs
        assert 'jayvdb/pywikibot-core/foo' not in slugs

    def test_targets(self):
        targets = get_stored_targets(self.base_dir, self.layout)

        assert targets
        assert type(targets) == list

        assert 'wikimedia/pywikibot-core/3052.11' in targets
        assert 'jayvdb/pywikibot-core/foo' not in targets

    def test_builds(self):
        builds = get_repo_stored_builds(
            self.base_dir, 'wikimedia/pywikibot-core', self.layout)

        assert builds
        assert 'wikimedia/pywikibot-core/3052' in builds


class TestAltLayout(object):

    base_dir = './tests/alt_layout'
    layout = '{job.state}/{job.repository.slug}/{job.number}.txt'

    def test_slugs(self):
        slugs = get_stored_repo_slugs(self.base_dir, self.layout)

        assert slugs == set(['a/b', 'foo/bar'])

    def test_builds(self):
        builds = get_repo_stored_builds(self.base_dir, 'a/b', self.layout)

        assert builds
        assert 'a/b/10' in builds

        builds = get_repo_stored_builds(self.base_dir, 'foo/bar', self.layout)

        assert builds
        assert 'foo/bar/7' in builds



"""Test Github resolution."""
from __future__ import absolute_import, unicode_literals

from travis_log_fetch.config import (
    _get_github,
    get_options,
)
from travis_log_fetch.get import (
    get_forks,
)

import pytest


# Note 'foo' is a real Github user, but they do not
# have repos bar or baz
class TestForks(object):

    def test_invalid(self):
        options = get_options()
        if not options.access_token:
            pytest.skip("github access token needed")
        _github = _get_github()
        pytest.raises(AssertionError, get_forks, _github, 'foo/bar')

    def test_zero(self):
        options = get_options()
        if not options.access_token:
            pytest.skip("github access token needed")
        _github = _get_github()
        forks = get_forks(_github, 'travispy/on_pypy')
        assert len(forks) == 0

    def test_fork(self):
        options = get_options()
        if not options.access_token:
            pytest.skip("github access token needed")
        _github = _get_github()
        forks = get_forks(_github, 'menegazzo/travispy')
        assert 'jayvdb/travispy' in forks



from __future__ import absolute_import, unicode_literals

import os

from travispy import TravisPy

GITHUB_ACCESS_TOKEN = os.environ.get('TRAVISPY_GITHUB_ACCESS_TOKEN')

TRAVIS_REPO_SLUG = os.environ.get('TRAVIS_REPO_SLUG')
TRAVIS_USER, TRAVIS_PROJECT = TRAVIS_REPO_SLUG.split('/', 1) if TRAVIS_REPO_SLUG else (None, None)


class LiveTest(object):

    def setup_method(self, method):
        if GITHUB_ACCESS_TOKEN:
            self._travis = TravisPy.github_auth(GITHUB_ACCESS_TOKEN)
        else:
            self._travis = TravisPy()



# -*- coding: utf-8 -*-
"""Command line handler."""
from __future__ import absolute_import
from __future__ import unicode_literals

import logging

from itertools import islice

from logging import getLogger

from travis_log_fetch import config

from travis_log_fetch._target import Target

from travis_log_fetch.get import (
    get_completed,
    get_forks,
    get_historical_builds,
    get_jobs,
    get_travis_repos,
    get_user_repos,
)
from travis_log_fetch._store import (
    download_job_log,
    get_stored_repo_slugs,
    skip_stored,
)


__logs__ = getLogger(__package__)


def main():
    """Main handler."""
    options = config.get_options()

    if not options or options.verbose:
        logging.basicConfig(
            level=logging.DEBUG,
            format='%(asctime)s %(levelname)9s %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S')

    __logs__.debug('{0!r}'.format(options))

    t = config._get_travispy()
    gh = config._get_github()
    if t:
        user = t.user()
    else:
        user = None

    targets = []

    for target_string in options.targets:
        if '://' in target_string:
            identifier = Target.from_url(target_string)
            targets.append(identifier)
        else:
            identifier = Target.from_extended_slug(target_string)
            targets.append(identifier)

    if options.refresh:
        slugs = get_stored_repo_slugs(options.dir, options.format)
        for slug in slugs:
            identifier = Target.from_simple_slug(slug)
            targets.append(identifier)

    if options.self:
        assert user
        targets += get_user_repos(t, user)

    if options.forks:
        for target in targets:
            slugs = get_forks(gh, target.slug)
            forks = get_travis_repos(t, slugs)
            targets += forks

    if options.all or options.old:
        count = None if options.all else options.count
        new_targets = []
        for target in targets:
            new_targets += list(islice(
                get_historical_builds(t, target.slug, _load_jobs=False),
                count))

        targets = new_targets

    # TODO: dont enumerate jobs if the files are all dated after the build end
    # TODO: enumerate all files starting with the job number,
    # and delete -started, -etc, when state is 'passed.

    if not options.force:
        targets = skip_stored(targets, options.dir, options.format)

    jobs = get_jobs(t, targets)

    if options.wait:
        jobs = get_completed(t, jobs, options.sleep)

    for job in jobs:
        download_job_log(options.dir, job, options.format)



"""Test Target resolution."""
from __future__ import absolute_import, unicode_literals

from travis_log_fetch.config import _get_travispy

from travis_log_fetch._target import Target
from travis_log_fetch.get import get_jobs

import pytest


# Note 'foo' is a real Github user, but they do not
# have repos bar or baz
class TestNonExistant(object):

    def test_single_target(self):
        target = Target.from_simple_slug('foo/bar')

        _travis = _get_travispy()
        jobs = get_jobs(_travis, target)
        assert len(jobs) == 0

    def test_multiple_targets(self):
        target1 = Target.from_simple_slug('foo/bar')
        target2 = Target.from_simple_slug('foo/baz')

        _travis = _get_travispy()
        jobs = get_jobs(_travis, [target1, target2])
        assert len(jobs) == 0


class TestLive(object):

    def test_single(self):
        target = Target.from_simple_slug('travispy/on_pypy')

        _travis = _get_travispy()
        jobs = get_jobs(_travis, target)
        assert len(jobs) == 1

        job = jobs[0]
        assert job.repository_id == 2598880

    def test_multiple(self):
        target1 = Target.from_simple_slug('travispy/on_pypy')
        target2 = Target.from_simple_slug('travispy/on_py34')

        _travis = _get_travispy()
        jobs = get_jobs(_travis, [target1, target2])
        assert len(jobs) == 2
        assert jobs[0].repository_id == 2598880
        assert jobs[1].repository_id == 2598879

    def test_logical_single_job_build(self):
        target = Target.from_extended_slug('travispy/on_pypy#1')

        _travis = _get_travispy()
        jobs = get_jobs(_travis, target)
        assert len(jobs) == 1

        job = jobs[0]
        assert job.repository_id == 2598880
        assert job.number == '1.1'
        assert job.id == 28881356

    def test_logical_multiple_job_build(self):
        target = Target.from_extended_slug('menegazzo/travispy#101')

        _travis = _get_travispy()
        jobs = get_jobs(_travis, target)

        assert len(jobs) == 5

        job = jobs[2]
        assert job.repository_id == 2419489
        assert job.number == '101.3'
        assert job.id == 82131391

    def test_logical_multiple_job_build_single_job(self):
        target = Target.from_extended_slug('menegazzo/travispy#101.3')

        _travis = _get_travispy()
        jobs = get_jobs(_travis, target)

        job = jobs[0]
        assert job.repository_id == 2419489
        assert job.number == '101.3'
        assert job.id == 82131391



# -*- coding: utf-8 -*-
"""Store logs."""
from __future__ import absolute_import
from __future__ import unicode_literals

import codecs
import datetime
import os

from logging import getLogger
from os.path import dirname, isdir

import dateutil
import dateutil.parser

import parse

import requests

from travis_log_fetch._target import Target

__logs__ = getLogger(__package__)

_HEADERS = {'Accept': 'text/plain; version=2'}


# TODO: add a 'clean' function to delete all logs which are
# associated with incomplete jobs.

# Create a new class to cache the stored metadata.


def get_files(base_dir):
    """Get list of all files under base_dir."""
    filenames = []
    for root, dirs, files in os.walk(base_dir):
        for file in files:
            filenames.append(os.path.join(root[len(base_dir):], file))

    return filenames


def get_stored_targets(base_dir, log_filename_format=None):
    """Get parsed stored targets."""
    assert log_filename_format
    if '{job.state}' in log_filename_format:
        log_filename_format = log_filename_format.replace(
            '{job.state}', '{job.state:w}')

    if '{job.number}' in log_filename_format:
        log_filename_format = log_filename_format.replace(
            '{job.number}', '{job.number:f}')

    parser = parse.compile(log_filename_format)

    if base_dir[-1] != '/':
        base_dir = base_dir + '/'

    filenames = get_files(base_dir)

    targets = []
    for filename in filenames:
        parsed_filename = parser.parse(filename)
        if not parsed_filename:
            __logs__.warning('Unexpected filename {0}'.format(filename))
            continue
        target = Target()
        try:
            target.slug = parsed_filename['job.repository.slug']
        except KeyError:
            pass
        try:
            target.number = str(parsed_filename['job.number'])
        except KeyError:
            pass
        try:
            target.state = parsed_filename['job.state']
        except KeyError:
            pass

        targets.append(target)

    return targets


def _select_repo_builds(targets, slug):
    """Filter targets for one repo."""
    targets = [target for target in targets if target.slug == slug]
    build_numbers = set(target.build_number for target in targets
                        if isinstance(target, Target))
    builds = [Target(slug=slug, build_number=build_number)
              for build_number in build_numbers]
    return builds


def get_repo_stored_builds(base_dir, slug, log_filename_format=None):
    """Get stored builds for a repo."""
    targets = get_stored_targets(base_dir, log_filename_format)
    builds = _select_repo_builds(targets, slug)
    return builds


def skip_stored(targets, base_dir, log_filename_format=None):
    """Optimistically skip targets that have been fetched."""
    assert log_filename_format
    stored_targets = get_stored_targets(base_dir, log_filename_format)

    new_targets = []
    for target in targets:
        if not isinstance(target, Target):
            target = Target._from_travispy_obj(target)

        slug = target.slug

        stored_repo_targets = _select_repo_builds(stored_targets, slug)
        if target not in stored_repo_targets:
            __logs__.debug('target {0} not found in {1!r}'.format(
                target, stored_repo_targets))
            new_targets.append(target)
        else:
            __logs__.info('skipping existing {0}'.format(target))

    return new_targets


def _get_simple_stored_repo_slugs(base_dir):
    """Get repo slugs from first two directory names."""
    if base_dir[-1] != '/':
        base_dir = base_dir + '/'

    slugs = [root[len(base_dir):] for root, dirs, files in os.walk(base_dir)
             if len(root[len(base_dir):].split(os.sep)) == 2]

    return slugs


def get_stored_repo_slugs(base_dir, log_filename_format=None):
    """Get existing directory slugs."""
    assert log_filename_format
    if log_filename_format.startswith('{job.repository.slug}/'):
        return _get_simple_stored_repo_slugs(base_dir)

    targets = get_stored_targets(base_dir, log_filename_format)
    slugs = set(target.slug for target in targets)
    return slugs


def download_job_log(base_dir, job, log_filename_format=None):
    """Download job log."""
    filename = log_filename_format.format(job=job)
    filename = '{0}/{1}'.format(base_dir, filename)

    if job.finished_at and os.path.exists(filename):
        file_ts = os.stat(filename).st_mtime
        file_ts = datetime.datetime.fromtimestamp(file_ts, dateutil.tz.tzutc())
        job_finish_ts = dateutil.parser.parse(job.finished_at)
        if file_ts >= job_finish_ts:
            return

    directory_name = dirname(filename)
    if not isdir(directory_name):
        os.makedirs(directory_name)

    text = job.log.body

    # FIXME(upstream): https://github.com/menegazzo/travispy/pull/27
    if not text:
        __logs__.info('fetching job {0} log directly'.format(job.id))
        r = requests.get('https://api.travis-ci.org/jobs/%s/log' % job.id,
                         headers=_HEADERS)
        text = r.content.decode('utf-8')

    try:
        with codecs.open(filename, 'w', 'utf8') as f:
            f.write(text)
    except UnicodeDecodeError as e:
        __logs__.warning(
            'UnicodeDecodeError while storing {0} into {1}: {2}'.format(
                type(text), filename, e))
        os.remove(filename)
        raise

    __logs__.info('wrote {0} with {1} chars'.format(filename, len(text)))



# -*- coding: utf-8 -*-
"""Setup module."""
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import sys

from setuptools import find_packages
from setuptools import setup

from setuptools.command.test import test as TestCommand  # noqa: N812


class PyTest(TestCommand):
    """Test harness."""

    user_options = []

    def initialize_options(self):
        """Initialise options hook."""
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        """Finalize options hook."""
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        """Run tests hook."""
        import pytest  # noqa: delayed import
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


_user = 'jayvdb'
_package_name = 'travis_log_fetch'

dependencies = [
    'travispy',
    'github3.py>=1.0.0a1',
    'parse',
    'python-dateutil',
    'ConfigArgParse',
]

dependency_links = [
    'git+https://github.com/jayvdb/travispy@fetch-log#egg=travispy-0.3.3'
]

_package_init_py = os.path.join(os.path.dirname(__file__),
                                _package_name,
                                '__init__.py')

with open(_package_init_py) as version_file:
    version_line = [line for line in version_file.readlines()
                    if line.startswith('__version__')][0]

version = version_line.split(' = ')[1][1:-2]

with open('README-pypi.rst') as readme_file:
    long_description = readme_file.read()

slug = _user + '/' + _package_name

setup(
    name=_package_name,
    packages=find_packages(),
    version=version,
    install_requires=dependencies,
    dependency_links=dependency_links,
    entry_points={
        'console_scripts': [
            'travis_log_fetch = travis_log_fetch:main',
        ],
    },
    description='Command line tool to fetch Travis-CI logs.',
    long_description=long_description,
    author='John Vandenberg',
    author_email='jayvdb@gmail.com',
    license='MIT',
    url='https://github.com/' + slug,
    download_url='https://github.com/' + slug + '/tarball/' + version,
    keywords=['travis', 'travis-ci', 'travispy'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing',
        'Topic :: Utilities'],
    # tests
    tests_require=['pytest'],
    cmdclass={'test': PyTest},
)



# -*- coding: utf-8 -*-
"""Get a travis log."""
from __future__ import absolute_import
from __future__ import unicode_literals

from logging import getLogger

from time import sleep

import travispy

from travis_log_fetch._target import Target

__logs__ = getLogger(__package__)


def get_travis_repo(t, slug):
    """
    Return Travis Repo or None.

    slugs without Travis builds emit a warning and return None.
    """
    try:
        return t.repo(slug)
    except travispy.errors.TravisError as e:
        __logs__.error('slug {0} not found on Travis: {1}'.format(slug, e))
        return None


def get_travis_repos(t, slugs):
    """
    Return list of Travis repos.

    slugs without Travis builds are omitted with a warning.
    """
    repos = []
    for slug in slugs:
        repo = get_travis_repo(t, slug)
        if repo:
            repos.append(repo)
    return repos


def get_jobs(t, targets):
    """Resolve targets to travis Jobs."""
    try:
        iter(targets)
    except TypeError:
        targets = [targets]

    jobs = []
    for target in targets:
        if isinstance(target, Target):
            repo = get_travis_repo(t, target.slug)
            if repo is None:
                continue
            if target.job_id:
                jobs = t.jobs(slug=repo.slug, ids=target.job_id)
                assert len(jobs) == 1
                target = jobs[0]

            elif target.build_id:
                target = t.build(target.build_id)
                assert target.repository_id == repo.id

            elif target.build_number:
                build = get_historical_build(t, target)
                if not target.job_number:
                    target = build
                else:
                    target = _get_build_job(t, build, target.job_number)

            else:
                target = repo

        if isinstance(target, travispy.Repo):
            if not repo.last_build_id:
                __logs__.error('No builds for {0}'.format(repo.slug))
                continue
            target = t.build(repo.last_build_id)

        if isinstance(target, travispy.Build):
            jobs += target.jobs
        elif isinstance(target, travispy.Job):
            jobs.append(target)
        else:
            raise AssertionError('Unexpected type: {0!r}'.format(target))

    return jobs


def get_completed(t, jobs, wait_time):
    """Return jobs as they are completed."""
    while jobs:
        completed = [job for job in jobs if not job.pending]
        for job in completed:
            yield job

        pending = [job for job in jobs if job.pending]
        if not pending:
            return

        __logs__.info('waiting {0} seconds for pending {1}'.format(
            wait_time, list(job.id for job in pending)))
        sleep(wait_time)
        jobs = [t.job(job.id) for job in pending]


def _fix_build_jobs(t, build):
    """Add the jobs attribute."""
    if not hasattr(build, 'jobs'):
        assert hasattr(build, 'job_ids')
        # FIXME(upstream): this doesnt work
        # build.jobs = t.jobs(ids=build.job_ids)
        build.jobs = [t.job(job_id) for job_id in build.job_ids]


def get_historical_builds(t, repo, _after=None, _load_jobs=True):
    """Get historical builds for a repo."""
    if isinstance(repo, travispy.entities.Repo):
        slug = repo.slug
    else:
        slug = repo
    builds = t.builds(slug=slug, after_number=_after)

    # As there can be duplicate build numbers, warn when
    # duplicates are encountered
    # https://github.com/travis-ci/travis-ci/issues/2582
    previous = None
    while builds:
        __logs__.debug('fetched {0} builds after {1}'.format(
            len(builds), _after))

        for build in builds:
            build_number = int(build.number)

            if _load_jobs:
                _fix_build_jobs(t, build)

            if previous and build_number == int(previous.number):
                __logs__.warning(
                    'Duplicate build {0} detected: {1} & {2}'.format(
                        build_number, build.id, previous.id))

            yield build

        _after = builds[-1].number
        builds = t.builds(slug=slug, after_number=_after)


def _get_build_job(t, build, job_number):
    """Get logical job from build."""
    _fix_build_jobs(t, build)
    if not job_number and len(build.jobs) == 1:
        return build.jobs[0]

    for build_job in build.jobs:
        _, _, build_job_number = build_job.number.partition('.')
        build_job_number = int(build_job_number)
        if build_job_number == job_number:
            return build_job

    raise AssertionError('{0}: could not find job {1}; max {2}'.format(
        build, job_number, build_job.number))


def get_historical_build(t, target):
    """Get historical build."""
    assert isinstance(target, Target)
    assert target.build_number
    repo = get_travis_repo(t, target.slug)
    after = target.build_number + 1
    builds = get_historical_builds(t, repo, _after=after)

    # As there can be duplicate build numbers, fetch one more
    # to detect duplicates
    # https://github.com/travis-ci/travis-ci/issues/2582
    found = None

    for build in builds:
        build_number = int(build.number)
        if build_number == target.build_number:
            if found:
                raise AssertionError('Duplicate build {0} detected'.format(
                    build_number))

            found = build

        if build_number < target.build_number:
            if found:
                return found
            break

    if found:
        return found

    raise AssertionError('could not find build {0}'.format(target))


def get_historical_job(t, target):
    """Get historical job."""
    assert isinstance(target, Target)
    assert target.job_number
    build = get_historical_build(t, target)
    return _get_build_job(t, build, target.job_number)


def get_user_repos(t, user):
    """Get user's travis repositories."""
    if hasattr(user, 'login'):
        user = user.login
    return t.repos(member=user)


def get_forks(gh, slug):
    """Get github fork slugs."""
    username, project = slug.rsplit('/', 1)
    repo = gh.repository(username, project)
    assert repo
    # github3 v1.0.0 has iter_forks, and 'forks' is count only
    if hasattr(repo, 'iter_forks'):
        forks = list(repo.iter_forks())
    else:
        forks = list(repo.forks())

    return [fork.full_name for fork in forks]
